import json
import io
import xml.etree.ElementTree as ET
from kmmarc.marc import Record, ControlField, DataField, SubField, Leader
from kmmarc.constants import *


class MarcJsonReader:
    def __init__(self, f) -> None:
        self.json = json.load(f)
        if not isinstance(self.json, list):
            self.json = [self.json]

    def __next(self, i):
        record_obj = self.json[i]
        record = Record(record_obj['leader'])

        if isinstance(record_obj['fields'], list):
            for field_obj in record_obj['fields']:
                tag = list(field_obj.keys())[0]

                if int(tag) < 10:
                    field = ControlField(tag, field_obj[tag])
                    record.control_fields.append(field)
                else:
                    field = DataField(tag, field_obj[tag]['ind1'], field_obj[tag]['ind2'])
                    
                    for subfield_obj in field_obj[tag]['subfields']:
                        code = list(subfield_obj.keys())[0]
                        field.subfields.append(SubField(code, subfield_obj[code]))

                    record.data_fields.append(field)
        else:
            for tag in record_obj['fields']:
                for field_obj in record_obj['fields']['tag']:
                    if int(tag) < 10:
                        field = ControlField(tag, field_obj)

                        record.control_fields.append(field)
                    else:
                        field = DataField(tag, field_obj['ind1'], field_obj['ind2'])

                        for code in field_obj['subfields']:
                            for subfield_obj in field_obj['subfields'][code]:
                                subfield = SubField(code, subfield_obj)
                                field.subfields.append(subfield)

                        record.data_fields.append(field)

        return record

    def __iter__(self):
        for i in range(len(self.json)):
            yield self.__next(i)


class MarcXmlReader:
    def __init__(self, data) -> None:
        self.root = data if isinstance(data, str) else ET.fromstringlist(data.readlines())
        self.record_tags = list(self.root)

    def __find(self, tag, path) -> ET.Element | None:
        res = tag.find(f"{'{http://www.loc.gov/MARC21/slim}'}{path}")
        if res is None:
            res = tag.find(f"{path}")
        return res
    
    def __findall(self, tag, path) -> list[ET.Element] :
        res = tag.findall(f"{'{http://www.loc.gov/MARC21/slim}'}{path}")
        if res is None or len(res) == 0:
            res = tag.findall(f"{path}")
        return res

    def __next(self, i):
        record_tag: ET.Element = self.record_tags[i]
        record = Record(self.__find(record_tag, "leader").text)

        for ctrl_field_tag in self.__findall(record_tag, "controlfield"):
            field = ControlField(ctrl_field_tag.attrib['tag'], ctrl_field_tag.text)
            record.control_fields.append(field)

        for data_field_tag in self.__findall(record_tag, "datafield"):
            field = DataField(data_field_tag.attrib['tag'], data_field_tag.attrib['ind1'], data_field_tag.attrib['ind2'])
            
            for subfield_tag in self.__findall(data_field_tag, "subfield"):
                subfield = SubField(subfield_tag.attrib['code'], subfield_tag.text)
                field.subfields.append(subfield)
            
            record.data_fields.append(field)

        return record

    def __iter__(self):
        for i in range(len(self.record_tags)):
            yield self.__next(i)


class MarcStreamReader:
    def __init__(self, f, force_utf8_encoding = False) -> None:
        self.__f = f
        self.__bytes: bytes = f.read()
        self.__bytes_len = len(self.__bytes) 
        self.__buf = io.BytesIO(self.__bytes)
        self.force_utf8_encoding = force_utf8_encoding

    def __parse_leader(self, leader_bytes: bytes):
        leader = Leader()
        
        leader_str = leader_bytes.decode("iso-8859-1")
        leader.record_length = int(leader_str[0:5])

        leader.record_status = leader_str[5:6]
        leader.type_of_record = leader_str[6:7]
        leader.impl_defined1 = leader_str[7:9]
        leader.char_coding_scheme = leader_str[9:10]
        leader.indicator_count = int(leader_str[10:11])
        leader.subfield_length = int(leader_str[11:12])
        leader.base_address_of_data = int(leader_str[12:17])
        leader.impl_defined2 = leader_str[17:20]
        leader.entry_map = leader_str[20:24]

        return leader
    
    def __parse_subfield_length(self, buf: io.BytesIO):
        global US
        global FT

        cur_bak = buf.tell()
        bytes_read = 0

        i = 0
        while True:
            if i >= 9999:
                buf.seek(cur_bak)
                return bytes_read

            r = buf.read(1)

            match r:
                case b'\x1f':
                    buf.seek(cur_bak)
                    return bytes_read
                case b'\x1e':
                    buf.seek(cur_bak)
                    return bytes_read
                case b'':
                    buf.seek(cur_bak)
                    raise Exception('Subfield not terminated')
                case _:
                    bytes_read += 1

            i += 1


    
    def __parse_data_field(self, tag, field_bytes: bytes, encoding: str):
        buf = io.BytesIO(field_bytes)

        field = DataField(tag, buf.read(1).decode(encoding), buf.read(1).decode(encoding))

        while True:
            read_byte = buf.read(1)
            if read_byte < b'\x00':
                break

            if read_byte == US:
                code = buf.read(1)
                if code < b'\x00':
                    raise Exception("Unexpected end of data field")
                
                if code == FT:
                    continue

                size = self.__parse_subfield_length(buf)
                data = buf.read(size)

                subfield = SubField(code.decode(encoding), data.decode(encoding))
                field.subfields.append(subfield)
                
                continue
            elif read_byte == FT:
                continue

        return field

    def __parse_record(self, leader_bytes: bytes, rec_bytes: bytes):
        leader = self.__parse_leader(leader_bytes)

        encoding = 'iso8859-1'
        if leader.char_coding_scheme == 'a' or self.force_utf8_encoding:
            encoding = 'utf-8'

        rec_buff = io.BytesIO(rec_bytes)

        directory_len = leader.base_address_of_data - (24 + 1)

        size = int(directory_len / 12)

        tags = [' '] * size
        lengths = [0] * size
        starts = [0] * size
        unsorted_start_index = {} 

        record = Record(leader)

        for i in range(size):
            tags[i] = rec_buff.read(3).decode("iso-8859-1")
            lengths[i] = int(rec_buff.read(4).decode("iso-8859-1"))
            starts[i] = int(rec_buff.read(5).decode("iso-8859-1"))
            unsorted_start_index[starts[i]] = i

        if rec_buff.read(1) != FT:
            raise Exception("Expected field terminator at end of directory")

        starts.sort()

        for s in range(size):
            i = unsorted_start_index[starts[s]]

            if int(tags[i]) < 10:
                eba = rec_buff.read(lengths[i] - 1)

                if rec_buff.read(1) != FT:
                    raise Exception("Expected field terminator at the end of field")

                record.control_fields.append(ControlField(tags[i], eba.decode(encoding)))
            else:
                eba = rec_buff.read(lengths[i])
                record.data_fields.append(self.__parse_data_field(tags[i], eba, encoding))
        
        if rec_buff.read(1) != RT:
            raise Exception("Expected record terminator at the end of record")

        return record

    def read_next(self):
        leader_bytes = self.__buf.read(24)

        rec_len = int(leader_bytes[0:5].decode("iso-8859-1"))
        return self.__parse_record(leader_bytes, self.__buf.read(rec_len - 24))

    def __iter__(self):
        while self.__buf.tell() < self.__bytes_len:
            yield self.read_next()



import json
import re

def parse_validations(builtin_valids = None):
    with open("validations.json", "r", encoding="utf-8") as f:
        jsondata = json.load(f)

    for key in builtin_valids:
        if key not in jsondata:
            jsondata[key] = builtin_valids[key]

    def process():
        for key in jsondata:
            obj = jsondata[key]
            if "$ref" in obj:
                ref = obj['$ref']
                jsondata[key] = jsondata[ref]

    process()
    
    def validate(code, value):
        obj = jsondata[code]
        if obj['type'] == 'enum':
            return obj['values'].count(value) > 0
        elif obj['type'] == 'regex':
            return len(re.findall(obj['value'], value)) != 0

    return lambda code, value : validate(code, value)

if __name__ == '__main__':
    valids = parse_validations({
        '200.c': {
            'type': 'enum',
            'values': ["JSON", "ISO"]
        }
    })
    print(valids('200.a', "EN"))
    print(valids('200.b', "123456789"))
    print(valids('200.d', "ES"))


import json
import yaml
import io
import xml.etree.ElementTree as ET
from kmmarc.marc import Record
from kmmarc.constants import *

class MarcJsonWriter:
    def __init__(self, f, layout_format: int = 1, ignored_tags: list[str] | None = None, indent: int | None = None):
        self.f = f
        self.format = layout_format
        self.ignored_tags = ignored_tags
        self.indent = indent

    def _write_format1(self, record: Record):
        obj = {
            'leader': str(record.leader),
            'fields': []
        }

        for field in record.control_fields:
            obj['fields'].append({
                field.tag: field.data
            })

        for field in record.data_fields:
            field_obj = {'ind1': field.ind1, 'ind2': field.ind2, 'subfields': []}

            for subfield in field.subfields:
                field_obj['subfields'].append({
                    subfield.code: subfield.data
                })

            obj['fields'].append({
                field.tag: field_obj
            })

        return obj

    def _write_format2(self, record: Record):
        obj = {
            'leader': record.leader,
            'fields': {}
        }

        for field in record.control_fields:
            if field.tag not in obj['fields']:
                obj['fields'][field.tag] = []

            obj['fields'][field.tag].append(field.data)

        for field in record.data_fields:
            if field.tag not in obj['fields']:
                obj['fields'][field.tag] = []

            field_obj = {'ind1': field.ind1, 'ind2': field.ind2, 'subfields': {}}

            for subfield in field.subfields:
                if subfield.code not in field_obj['subfields']:
                    field_obj['subfields'][subfield.code] = []

                field_obj['subfields'][subfield.code].append(subfield.data)

            obj['fields'][field.tag].append(field_obj)

        return obj

    def write(self, record: Record):
        if self.format == 1:
            obj = self._write_format1(record)
        else:
            obj = self._write_format2(record)

        json.dump(obj, self.f, indent=self.indent)

    def write_all(self, records: list[Record]):
        recs = []
        for record in records:
            if self.format == 1:
                recs.append(self._write_format1(record))
            else:
                recs.append(self._write_format2(record))

        json.dump(recs, self.f, indent=self.indent)


class MarcYamlWriter(MarcJsonWriter):
    def write(self, record: Record):
        if self.format == 1:
            obj = self._write_format1(record)
        else:
            obj = self._write_format2(record)

        yaml.dump(obj, self.f, indent=self.indent, sort_keys=False)

    def write_all(self, records: list[Record]):
        recs = []
        for record in records:
            if self.format == 1:
                recs.append(self._write_format1(record))
            else:
                recs.append(self._write_format2(record))

        yaml.dump(recs, self.f, indent=self.indent, sort_keys=False)


class MarcXmlWriter:
    def __init__(self, f, indent: int | None = None, xml_declaration=True, use_marc_namespace=False) -> None:
        self.f = f
        self.xml_declaration = xml_declaration
        self.indent = indent
        self.namespace = 'marc:' if use_marc_namespace else ''
        self.collection_tag = ET.Element(f'{self.namespace}collection')
        if use_marc_namespace:
            self.collection_tag.attrib['xmlns:marc'] = 'http://www.loc.gov/MARC21/slim'

    def write(self, record: Record):
        record_tag = ET.SubElement(self.collection_tag, 'record')
        leader_tag = ET.SubElement(record_tag, f'{self.namespace}leader')
        leader_tag.text = record.leader

        for field in record.control_fields:
            field_tag = ET.SubElement(record_tag, f'{self.namespace}controlfield')
            field_tag.attrib['tag'] = field.tag
            field_tag.text = field.data

        for field in record.data_fields:
            field_tag = ET.SubElement(record_tag, f'{self.namespace}datafield')
            field_tag.attrib['tag'] = field.tag
            field_tag.attrib['ind1'] = field.ind1
            field_tag.attrib['ind2'] = field.ind2

            for subfield in field.subfields:
                subfield_tag = ET.SubElement(field_tag, f'{self.namespace}subfield')
                subfield_tag.attrib['code'] = subfield.code
                subfield_tag.text = subfield.data

    def write_all(self, *records):
        for record in records:
            self.write(record)

    def flush(self):
        if self.indent is not None:
            ET.indent(self.collection_tag, space=''.join([" "] * self.indent))

        self.f.write(ET.tostring(self.collection_tag, xml_declaration=self.xml_declaration, encoding="unicode"))


class MarcStreamWriter:
    def __init__(self, f: io.FileIO, force_utf8_encoding=False) -> None:
        self.f = f
        self.force_utf8_encoding = force_utf8_encoding

    def write(self, record: Record):
        dir_buf = io.BytesIO()
        data_buf = io.BytesIO()

        previous = 0

        ldr = record.leader
        encoding = "iso8859-1"
        if self.force_utf8_encoding:
            ldr.char_coding_scheme = "a"
            encoding = 'utf-8'

        for field in record.control_fields:
            data_buf.write(field.data.encode(encoding))
            data_buf.write(FT)
            dir_buf.write(f"{field.tag}{(data_buf.tell() - previous):04d}{previous:05d}".encode('iso8859-1'))
            previous = data_buf.tell()

        for field in record.data_fields:
            data_buf.write(field.ind1.encode(encoding))
            data_buf.write(field.ind2.encode(encoding))
            for subfield in field.subfields:
                data_buf.write(US)
                data_buf.write(subfield.code.encode(encoding))
                data_buf.write(subfield.data.encode(encoding))
            data_buf.write(FT)
            dir_buf.write(f"{field.tag}{(data_buf.tell() - previous):04d}{previous:05d}".encode('iso8859-1'))
            previous = data_buf.tell()

        dir_buf.write(FT)

        base_address = 24 + dir_buf.tell()
        ldr.base_address_of_data = base_address
        record_len = ldr.base_address_of_data + data_buf.tell() + 1
        ldr.record_length = record_len

        self.f.write(f"{ldr.record_length:05d}".encode('iso8859-1'))
        self.f.write(f"{ldr.record_status}".encode('iso8859-1'))
        self.f.write(f"{ldr.type_of_record}".encode('iso8859-1'))
        self.f.write(f"{''.join(ldr.impl_defined1)}".encode('iso8859-1'))
        self.f.write(f"{ldr.char_coding_scheme}".encode('iso8859-1'))
        self.f.write(f"{str(ldr.indicator_count)}".encode('iso8859-1'))
        self.f.write(f"{str(ldr.subfield_length)}".encode('iso8859-1'))
        self.f.write(f"{ldr.base_address_of_data:05d}".encode('iso8859-1'))
        self.f.write(f"{''.join(ldr.impl_defined2)}".encode('iso8859-1'))
        self.f.write(f"{''.join(ldr.entry_map)}".encode('iso8859-1'))

        dir_buf.seek(0)
        data_buf.seek(0)
        self.f.write(dir_buf.read())
        self.f.write(data_buf.read())
        self.f.write(RT)

    def write_all(self, *records):
        for record in records:
            self.write(record)



from kmmarc.marc import *

def format_record_as_unimarc_view(record: Record):
    res = f"MFN: 0\nEstado: UNK   Tipo: UNK   Nível hierárquico: UNK  Nível de cod: UNK\n"
    for ctrl_field in sorted(record.control_fields, key=lambda cs : cs.tag):
        res += f"\n{ctrl_field.tag}:{ctrl_field.data}"
    for data_field in sorted(record.data_fields, key=lambda sf : sf.tag):
        res += f"\n{data_field.tag}:{data_field.ind1}{data_field.ind2}"

        for subfield in data_field.subfields:
            res += f"^{subfield.code}{subfield.data}"
    return res


def format_record_as_isbd_view(record: Record):
    res = "[X]\n\n"
    res += f"{record['700'][0]['a'][0].data}, {record['700'][0]['b'][0].data}\n"
    res += f"{record['Title']} / {record['200'][0]['f'][0].data}"
    res += "\n\nCDU: 00"
    return res


class VariableField:
    def __init__(self, tag: str) -> None:
        self.tag = tag

class ControlField(VariableField):
    def __init__(self, tag: str, data: str) -> None:
        super().__init__(tag)
        self.data = data

    def __str__(self) -> str:
        return f"{self.tag} {self.data}"

class SubField:
    def __init__(self, code: str, data: str) -> None:
        self.code = code
        self.data = data

    def __str__(self) -> str:
        return f"${self.code}{self.data}"


class DataField(VariableField):
    def __init__(self, tag: str, ind1: str, ind2: str) -> None:
        super().__init__(tag)
        self.ind1 = ind1
        self.ind2 = ind2
        self.subfields: list[SubField] = []

    def __getitem__(self, key) -> list[SubField] | None:
        res = []
        
        for subfield in self.subfields:
            if subfield.code == key:
                res.append(subfield)
        
        return res if len(res) > 0 else None 

    def __contains__(self, key) -> bool:
        for subfield in self.subfields:
            if subfield.code == key:
                return True

        return False

    def __str__(self) -> str:
        res = f"{self.tag} {self.ind1}{self.ind2}"
        for subfield in self.subfields:
            res += str(subfield)
        return res


class Leader:
    def __init__(self) -> None:
        self.record_length = 0
        self.record_status: str = ' '
        self.type_of_record: str = ' '
        self.impl_defined1 = []
        self.char_coding_scheme = []
        self.indicator_count = 0
        self.subfield_length = 0
        self.base_address_of_data = 0
        self.impl_defined2 = []
        self.entry_map: list[str] = []

    def __str__(self) -> str:
        return f"=LDR {self.record_length:05d}{self.record_status}{self.type_of_record}{''.join(self.impl_defined1)}{self.char_coding_scheme}{self.indicator_count}{self.subfield_length}{self.base_address_of_data:05d}{''.join(self.impl_defined2)}{''.join(self.entry_map)}"


class Record:
    def __init__(self, leader: Leader) -> None:
        self.leader = leader
        self.control_fields: list[ControlField] = []
        self.data_fields: list[DataField] = []

    def __getitem__(self, key) -> list[ControlField | DataField] | None:
        if key in Record.__custom_getters:
            try:
                return Record.__custom_getters[key](self)
            except:
                return None

        res = []
        
        for field in self.control_fields:
            if field.tag == key:
                res.append(field)

        for field in self.data_fields:
            if field.tag == key:
                res.append(field)
        
        return res if len(res) > 0 else None 
    
    def __contains__(self, key) -> bool:
        for field in self.control_fields:
            if field.tag == key:
                return True
            
        for field in self.data_fields:
            if field.tag == key:
                return True

        return False

    def __str__(self) -> str:
        res = f"{self.leader}"
        for field in sorted(self.control_fields, key=lambda cs : cs.tag):
            res += f"\n={field}"
        for field in sorted(self.data_fields, key=lambda sf : sf.tag):
            res += f"\n={field}"
        return res
    
    __custom_getters = {}
    
    @staticmethod
    def register_custom_getter(name: str, getter):
        Record.__custom_getters[name] = getter
